import struct

from hks_pylib.cryptography.ciphers.hkscipher import HKSCipher
from hks_pylib.cryptography.ciphers.cipherid import CipherID, hash_cls_name

from hks_pynetwork.packet import MIN_HEADER_SIZE, PacketEncoder, PacketDecoder

from hks_pynetwork.errors.secure_packet import CipherTypeMismatchError
 

class SecurePacketEncoder(PacketEncoder):
    def __init__(self, cipher: HKSCipher):
        self.cipher = cipher

    def encode(self, payload: bytes):
        payload = self.cipher.encrypt(payload)
        packet = super().encode(payload)

        # SECURE HEADER = TYPE_OF_CIPHER (2 bytes) + NUMBER_OF_PARAMS(1 byte)
        #                 + PARAM1_SIZE + PARAM1 + PARAM2_SIZE + PARAM2 + ...
        # TYPE_OF_CIPHER is the hash value of the cipher class

        secure_header = hash_cls_name(self.cipher)\
            + struct.pack(">B", self.cipher._number_of_params)

        for i in range(self.cipher._number_of_params):
            param = self.cipher.get_param(i)

            assert isinstance(param, bytes)

            param_size = len(param)
            param_struct = "B{}s".format(param_size)
            secure_header += struct.pack(param_struct, param_size, param)

        # Set new header size = old header size + secure header size
        old_header_size = struct.unpack(">H", packet[:2])[0]
        
        packet = packet[:old_header_size]\
            + secure_header\
            + packet[old_header_size:]
        
        new_header_size = old_header_size + len(secure_header)
        new_header_size = struct.pack(">H", new_header_size)

        # Return the packet with new header size
        return new_header_size + packet[2:]


class SecurePacketDecoder(PacketDecoder):
    def __init__(self, cipher: HKSCipher):
        self.cipher = cipher

    def decode(self, packet: bytes):
        packet_dict = super().decode(packet)

        # ORIGINAL HEADER = HEADER_SIZE (2 bytes) + PAYLOAD_SIZE (2 byte)

        # SECURE HEADER: TYPE_OF_CIPHER (2 bytes) + NUMBER_OF_PARAMS(1 byte)
        #                 + PARAM1_SIZE + PARAM1 + PARAM2_SIZE + PARAM2 + ...

        cipher_hashvalue = packet[MIN_HEADER_SIZE: MIN_HEADER_SIZE + 2]
        cipher_type = CipherID.hash2cls(cipher_hashvalue)

        if cipher_type is None:
            raise CipherTypeMismatchError("Cipher type is invalid "
            "(hash = {}).".format(cipher_hashvalue))

        if not isinstance(self.cipher, cipher_type):
            raise CipherTypeMismatchError("Cipher type mismatches "
            "(expected {}, but received {}).".format(
                type(self.cipher).__name__,
                cipher_type.__name__
            ))

        number_of_params = struct.unpack(
                ">B",
                packet[MIN_HEADER_SIZE + 2: MIN_HEADER_SIZE + 3]
            )[0]

        current_index = MIN_HEADER_SIZE + 3
        for i in range(number_of_params):
            param_size = struct.unpack(
                    ">B",
                    packet[current_index: current_index + 1]
                )[0]
            current_index += 1

            param = struct.unpack(
                    ">{}s".format(param_size),
                    packet[current_index: current_index + param_size]
                )[0]
            current_index += param_size

            self.cipher.set_param(i, param)

        self.cipher.reset(False)

        packet_dict["payload"] = self.cipher.decrypt(packet_dict["payload"])

        return packet_dict



import copy
import time
import errno
import socket
import threading

from hks_pylib.logger import LoggerGenerator
from hks_pylib.cryptography.ciphers.cipherid import CipherID
from hks_pylib.cryptography.ciphers.hkscipher import HKSCipher

from hks_pynetwork.packet_buffer import PacketBuffer
from hks_pynetwork.secure_packet import SecurePacketEncoder, SecurePacketDecoder

from hks_pylib.errors.cryptography.ciphers import InvalidCipherParameterError
from hks_pylib.errors.cryptography.ciphers import NotEnoughCipherParameterError
from hks_pylib.errors.cryptography.ciphers.symmetrics import UnAuthenticatedPacketError

from hks_pynetwork.errors.packet import IncompletePacketError
from hks_pynetwork.errors.secure_packet import CipherTypeMismatchError
from hks_pynetwork.errors.external import STCPSocketClosedError, STCPSocketTimeoutError


class STCPSocket(object):
    DEFAULT_TIME_OUT = 0.1
    DEFAULT_RELOAD_TIME = 0.1

    def __init__(
                    self,
                    cipher: HKSCipher,
                    name: str,
                    buffer_size: int,
                    logger_generator: LoggerGenerator,
                    display: dict
                ):
        assert isinstance(cipher, HKSCipher)
        assert isinstance(buffer_size, int)
        assert isinstance(logger_generator, LoggerGenerator)
        assert isinstance(display, dict)
        assert isinstance(name, str)

        self._logger_generator = logger_generator
        self._display = display
        self._log = self._logger_generator.generate(name, self._display)

        self._log("dev", "debug", "Initialized with cipher {}.".format(
            CipherID.cls2name(type(cipher))
            ))
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__cipher = cipher
        self.__cipher.reset()

        self.set_reload_time(STCPSocket.DEFAULT_RELOAD_TIME)
        self.set_recv_timeout(None)

        self.__packet_encoder = SecurePacketEncoder(self.__cipher)
        self.__packet_decoder = SecurePacketDecoder(self.__cipher)
        self.__buffer = None
        self.__buffer_size = buffer_size

        self.__buffer_available = threading.Event()
        self._stop_auto_recv = False
        self._prepare_close = False

        self._is_working = False

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self._is_working:
            self.close()

    def _start_auto_recv(self):
        self._log("dev", "info", "Start automatically receiving...")
        while True:
            try:
                data = self._socket.recv(self.__buffer_size)
            except socket.error as e:
                if isinstance(e, socket.timeout) and not self._stop_auto_recv:
                    continue
 
                self._socket.close()
                if e.errno in (errno.ECONNRESET, errno.ECONNABORTED, errno.ECONNREFUSED):
                    self._log("dev", "info", "Automatic receiving closes "
                    "normally ({}).".format(repr(e)))
                    break
                elif isinstance(e, socket.timeout):
                    self._log("dev", "info", "Automatic receiving closes "
                    "normally (timeout and socket closed).")
                    break
                elif isinstance(e, OSError) and e.errno == 10038:
                    self._log("dev", "info", "Automatic receiving closes "
                    "normally (remote socket closed).")
                    break
                else:
                    self._log("dev", "error", "Socket error is"
                    " occured with a unknown reason ({}).".format(repr(e)))
                    raise e
            except Exception as e:
                self._socket.close()
                self._log("dev", "error", "Unknown error in automatically "
                "receiving ({}).".format(repr(e)))
                raise e
            else:
                # when be closed, socket will receive infinite empty packets
                if not data:
                    self._socket.close()
                    self._log("dev", "info", "Automatic receiving closes "
                    "normally (remote socket closed).")
                    break
                self.__buffer.push(data)
                self.__buffer_available.set()

        self.__buffer_available.set()
        self._log("dev", "info", "Stop automatic receiving...")

    def set_reload_time(self, value: float):
        assert value >= 0

        self._log("dev", "debug", "Set reload time to {}.".format(value))
        self.__reload_time = value

    def set_recv_timeout(self, value: float):
        assert value is None or value >= 0

        self._log("dev", "debug", "Set receive timeout to {}.".format(value))
        self.__recv_timeout = value

    def settimeout_raw(self, value: float):
        self._log("dev", "debug", "Set timeout to {}.".format(value))
        return self._socket.settimeout(value)

    def recv(self) -> bytes:
        data = b''
        if self.__recv_timeout is not None:
            start_recv_time = time.time()

        while not data:
            if self.isclosed() and len(self.__buffer) == 0:
                raise STCPSocketClosedError("Connection closed.")

            try:
                data = self.__buffer.pop()
            except IncompletePacketError as e:
                # Not enough length of packet, waiting more data
                continue
            except CipherTypeMismatchError as e:
                self._log("user", "warning", "Detect an abnormal packet.")
                self._log("dev", "warning", "Detect an abnormal packet "
                "({}).".format(repr(e)))
                break
            except (
                    UnAuthenticatedPacketError,
                    InvalidCipherParameterError,
                    NotEnoughCipherParameterError
                ) as e:
                # Cannot decrypt the packet: authentication failed, wrong parameters, ...
                self._log("user", "warning", "Detect an abnormal packet.")
                self._log("dev", "warning", "Detect an abnormal packet "
                "({}).".format(repr(e)))
                break
            except Exception as e:
                self._log("user", "info", "Unknown error in recv() method.")
                self._log("dev", "error", "Unknown error in recv() method "
                "({}).".format(repr(e)))
                break
            finally:
                if not data:
                    if self.__recv_timeout is not None and\
                        time.time() - start_recv_time > self.__recv_timeout:
                        if self.__recv_timeout != 0:
                            raise STCPSocketTimeoutError("Receiving exceeds timeout.")
                        else:
                            raise STCPSocketTimeoutError("Method recv() can not return "
                            "the value immediately.")

                    time.sleep(self.__reload_time)

                    if self.__recv_timeout is None and len(self.__buffer) == 0\
                        and not self.isclosed():
                        self.__buffer_available.wait()

        self.__buffer_available.clear()
        return data

    def send(self, data) -> int:
        self.__cipher.reset()
        packet = self.__packet_encoder.encode(data)
        return self._socket.send(packet)

    def sendall(self, data) -> int:
        self.__cipher.reset()
        packet = self.__packet_encoder.encode(data)
        return self._socket.sendall(packet)

    def bind(self, address):
        return self._socket.bind(address)

    def listen(self, __backlog: int = 0):
        self._log("user", "info", "Server listen...")
        self._log("dev", "info", "Server listen...")
        self._socket.listen(__backlog)
        self._is_working = True

    def accept(self):
        socket, addr = self._socket.accept()
        socket.settimeout(STCPSocket.DEFAULT_TIME_OUT)  # timeout for non-blocking socket
        
        self._log("user", "info", "Server accept {}.".format(addr))
        self._log("dev", "info", "Server accept {}.".format(addr))
        
        socket = self._fromsocket(socket, addr, start_serve=True)
        socket._is_working = True
        
        return socket, addr

    def connect(self, address):
        ret = self._socket.connect(address)
        self._is_working = True

        self._log("user", "info", "Connect to server {} successfully.".format(address))
        self._log("dev", "info", "Connect to server {} successfully.".format(address))
        self._log("dev", "debug", "Transform to STCP Socket {}.".format(address))

        self._log = self._logger_generator.generate(
                f"STCP Socket {address}", self._display
            )
        self.settimeout_raw(STCPSocket.DEFAULT_TIME_OUT)

        self._stop_auto_recv = False
        server = threading.Thread(target=self._start_auto_recv)
        server.start()

        self.__buffer = PacketBuffer(
                decoder=self.__packet_decoder,
                name="Packet Buffer of {}".format(address),
                logger_generator=self._logger_generator,
                display=self._display
            )
        return ret

    def close(self):
        if self.__buffer is not None:  # not STCP Listener
            self.__buffer_available.set()
            self._stop_auto_recv = True

        if not self.isclosed():
            self._socket.close()
            self._is_working = False
            self._log("dev", "info", "Closed.")

    def isclosed(self):
        "This method returns the raw socket._closed"
        return self._socket._closed

    def isworking(self):
        return self._is_working and not self.isclosed()

    def _fromsocket(self, socket: socket.socket, address, start_serve=True):
        cipher = copy.copy(self.__cipher)

        new_socket = STCPSocket(
                cipher=cipher, 
                buffer_size=self.__buffer_size,
                logger_generator=self._logger_generator,
                display=self._display,
                name=f"STCP Socket {address}"
            )

        new_socket._socket = socket

        new_socket.__buffer = PacketBuffer(
                decoder=new_socket.__packet_decoder,
                name="PacketBuffer of {}".format(address),
                logger_generator=self._logger_generator,
                display=self._display
            )

        if start_serve:
            new_socket._stop_auto_recv = False
            server = threading.Thread(target=new_socket._start_auto_recv)
            server.start()
        return new_socket



from hks_pynetwork.errors import HKSPyNetworkError


class InternalError(HKSPyNetworkError):
    "Exception is raised by failures in internal module."


class ChannelSlotError(InternalError):
    "Exception is raised when an error occurs due to slot."


class ChannelClosedError(InternalError):
    "Exception is raised when the method is working after channel closing."



from hks_pynetwork.errors import HKSPyNetworkError


class PacketError(HKSPyNetworkError):
    "Exception is raised by failures in packet module."


class PacketSizeError(PacketError):
    "Exception raised by the invalid size of packet elements"
    

class CannotExtractPacketError(PacketError):
    "Exception is raised by some errors in packet extracting."


class IncompletePacketError(CannotExtractPacketError):
    "Exception is raised when the incomplete packet is extracted."



import threading

from hks_pylib.logger import LoggerGenerator
from hks_pynetwork.secure_packet import PacketDecoder

from hks_pynetwork.errors.packet import IncompletePacketError
from hks_pynetwork.errors.secure_packet import CipherTypeMismatchError


class PacketBuffer():
    def __init__(
                    self,
                    decoder: PacketDecoder,
                    name: str,
                    logger_generator: LoggerGenerator,
                    display: dict
                ) -> None:
        assert isinstance(decoder, PacketDecoder)

        self._buffer = []

        self._packet_decoder = decoder

        self.__print = logger_generator.generate(name, display)
        
        self._current_packet = b""
        self._current_packet_size = 0
        self._expected_current_packet_size = 0
        
        self._push_lock = threading.Lock()

    def push(self, packet: bytes, append_to_end: bool = True):
        self._push_lock.acquire()

        if append_to_end:
            self._buffer.append(packet)
        else:
            self._buffer.insert(0, packet)

        self._push_lock.release()

    def pop(self):
        if len(self._buffer) == 0:
            return b""

        self._current_packet_size += len(self._buffer[0])
        self._current_packet += self._buffer[0]
        del self._buffer[0]

        if self._packet_decoder is None:
            ret = self._current_packet
            self._current_packet = b""
            self._current_packet_size = 0
            self._expected_current_packet_size = 0
            return ret

        if self._expected_current_packet_size == 0:
            try:
                packet_dict = self._packet_decoder.get_header(self._current_packet)
                self._expected_current_packet_size =\
                    packet_dict["payload_size"] + packet_dict["header_size"]
            except IncompletePacketError:
                return b""
            except Exception as e:
                self.__print("dev", "error", "Unknown error occurs "
                "when decode a packet ({}).".format(repr(e)))
                raise e

        if self._current_packet_size < self._expected_current_packet_size:
            return b""

        try:
            packet_dict = self._packet_decoder.decode(self._current_packet)
        except CipherTypeMismatchError as e:
            raise e
        finally:
            if  self._current_packet_size > self._expected_current_packet_size:
                apart_of_next_packet = self._current_packet[
                        self._expected_current_packet_size :
                    ]
                self.push(apart_of_next_packet, append_to_end=False)
 
            self._current_packet = b""
            self._current_packet_size = 0
            self._expected_current_packet_size = 0

        return packet_dict["payload"]

    def __len__(self):
        return len(self._buffer)



from hks_pynetwork.errors import HKSPyNetworkError


class SecurePacketError(HKSPyNetworkError):
    "Exception is raised by failures in secure_packet module."


class CipherTypeMismatchError(SecurePacketError):
    """Exception is raised when type of cipher of
    decoded packet is diffrent from intial cipher."""



import struct

from hks_pynetwork.errors.packet import IncompletePacketError, PacketSizeError


MAX_HEADER_SIZE = 65535
MAX_PAYLOAD_SIZE = 4294967295
MAX_PACKET_SIZE = MAX_HEADER_SIZE + MAX_PAYLOAD_SIZE

MIN_HEADER_SIZE = 6


class PacketEncoder(object):
    def encode(self, payload: bytes):
        assert isinstance(payload, bytes)

        if len(payload) > MAX_PAYLOAD_SIZE:
            raise PacketSizeError("Payload size is too large "
                "(expected < {}).".format(MAX_PAYLOAD_SIZE))

        # PACKET = HEADER + PAYLOAD
        # HEADER = HEADER_SIZE(2 bytes) + PAYLOAD_SIZE(4 byte) + OPTIONAL_HEADER
        # ==> PACKET = HERDER_SIZE + PAYLOAD_SIZE + OPTIONAL_HEADER + PAYLOAD

        header_struct = ">HI"

        header_dummy = struct.pack(header_struct, 0, 0)
        header_size = len(header_dummy)

        header = struct.pack(header_struct, header_size, len(payload))

        return header + payload


class PacketDecoder(object):
    def get_header(self, packet: bytes):
        assert isinstance(packet, bytes)

        if len(packet) > MAX_PACKET_SIZE:
            raise PacketSizeError("Packet size is too large "
                "(expected < {}).".format(MAX_PACKET_SIZE))

        if len(packet) < MIN_HEADER_SIZE:
            raise IncompletePacketError("Incomplete header.")

        header_size, payload_size = struct.unpack(">HI", packet[:MIN_HEADER_SIZE])
        if header_size > MAX_HEADER_SIZE:
            raise PacketSizeError("Header size is too large "
                "(expected < {}).".format(MAX_HEADER_SIZE))

        if len(packet) < header_size:
            raise IncompletePacketError("Incomplete header.")

        header_dict = {}
        header_dict["header_size"] = header_size
        header_dict["payload_size"] = payload_size

        return header_dict

    def decode(self, packet: bytes):
        assert isinstance(packet, bytes)

        header = self.get_header(packet)

        header_size, payload_size = header["header_size"], header["payload_size"]
        
        if payload_size > MAX_PAYLOAD_SIZE:
            raise PacketSizeError("Payload size is too large "
                "(expected < {}).".format(MAX_PAYLOAD_SIZE))

        if len(packet) < header_size + payload_size:
            raise IncompletePacketError("Incomplete packet.")
 
        packet_dict = header
        packet_dict["payload"] = packet[header_size: header_size + payload_size]
        packet_dict["packet_size"] = header_size + payload_size

        return packet_dict



import setuptools


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
    long_description += "\n"


with open("CHANGELOG.md", "r", encoding="utf-8") as fh:
    long_description += fh.read()
 

setuptools.setup(
    name="hks_pynetwork",
    version="0.0.3",
    author="huykingsofm",
    author_email="huykingsofm@gmail.com",
    description="A Python library is built to communicate between objects in internal or external programs.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/huykingsofm/hks_pynetwork",
    project_urls={
        "Bug Tracker": "https://github.com/huykingsofm/hks_pynetwork/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.7.1",
    install_requires=["hks_pylib==0.0.5"],
    setup_requires=["pytest-runner==4.4"],
    tests_require=["pytest==4.4.1", "pytest-parallel==0.1.0"],
    test_suite="tests",
)


from hks_pynetwork.errors import HKSPyNetworkError


class ExternalError(HKSPyNetworkError):
    "Exception is raised by failures in external module."


class STCPSocketClosedError(ExternalError):
    "Exception is raised when a STCPSocket was closed."


class STCPSocketTimeoutError(ExternalError):
    "Exception is raised when a STCPSocket method stop due to timeout."
